@@ -1,7 +1,7 @@
 package Mail::Milter::Authentication::Handler;
 use strict;
 use warnings;
-our $VERSION = '2.20181130'; # VERSION
+our $VERSION = '2.20181212'; # VERSION
 
 
 use Digest::MD5 qw{ md5_hex };
@@ -368,6 +368,19 @@ sub _remap_ip_and_helo {
 }
 
 
+sub remap_connect_callback {
+    my ( $self, $hostname, $ip ) = @_;
+    $self->{'raw_ip_object'} = $ip;
+    my $ip_remap = $self->_remap_ip_and_helo();
+    if ( $ip_remap ) {
+        $ip = $ip_remap->{ip};
+        $self->dbgout( 'RemappedConnect', $self->{'raw_ip_object'}->ip() . ' > ' . $ip->ip(), LOG_DEBUG );
+    }
+    $self->{'ip_object'} = $ip;
+    return;
+}
+
+
 sub top_connect_callback {
 
     # On Connect
@@ -387,15 +400,6 @@ sub top_connect_callback {
         }
 
         $self->dbgout( 'ConnectFrom', $ip->ip(), LOG_DEBUG );
-        $self->{'raw_ip_object'} = $ip;
-
-        my $ip_remap = $self->_remap_ip_and_helo();
-        if ( $ip_remap ) {
-            $ip = $ip_remap->{ip};
-            $self->dbgout( 'ConnectFromRemapped', $self->{'raw_ip_object'}->ip() . ' > ' . $ip->ip(), LOG_DEBUG );
-        }
-
-        $self->{'ip_object'} = $ip;
 
         my $callbacks = $self->get_callbacks( 'connect' );
         foreach my $handler ( @$callbacks ) {
@@ -431,6 +435,29 @@ sub top_connect_callback {
 }
 
 
+sub remap_helo_callback {
+    my ( $self, $helo_host ) = @_;
+    if ( !( $self->{'helo_name'} ) ) {
+
+        $self->{'raw_helo_name'} = $helo_host;
+        my $ip_remap = $self->_remap_ip_and_helo();
+        if ( $ip_remap ) {
+            my $ip = $ip_remap->{ip};
+            if ( $self->{'ip_object'}->ip() ne $ip_remap->{ip}->ip() ) {
+                # The mapped IP has been changed based on the HELO host received
+                $self->{'ip_object'} = $ip;
+                $self->dbgout( 'RemappedConnectHELO', $self->{'ip_object'}->ip() . ' > ' . $ip->ip(), LOG_DEBUG );
+            }
+            $helo_host = $ip_remap->{helo};
+            $self->dbgout( 'RemappedHELO', $self->{'raw_helo_name'} . ' > ' . $helo_host, LOG_DEBUG );
+        }
+
+        $self->{'helo_name'} = $helo_host;
+    }
+    return;
+}
+
+
 sub top_helo_callback {
 
     # On HELO
@@ -447,22 +474,8 @@ sub top_helo_callback {
         }
 
         # Take only the first HELO from a connection
-        if ( !( $self->{'helo_name'} ) ) {
-
-            $self->{'raw_helo_name'} = $helo_host;
-            my $ip_remap = $self->_remap_ip_and_helo();
-            if ( $ip_remap ) {
-                my $ip = $ip_remap->{ip};
-                if ( $self->{'ip_object'}->ip() ne $ip_remap->{ip}->ip() ) {
-                    # The mapped IP has been changed based on the HELO host received
-                    $self->{'ip_object'} = $ip;
-                    $self->dbgout( 'ConnectFromRemappedAtHELO', $self->{'ip_object'}->ip() . ' > ' . $ip->ip(), LOG_DEBUG );
-                }
-                $helo_host = $ip_remap->{helo};
-                $self->dbgout( 'HELORemapped', $self->{'raw_helo_name'} . ' > ' . $helo_host, LOG_DEBUG );
-            }
-
-            $self->{'helo_name'} = $helo_host;
+        if ( !( $self->{'seen_helo_name'} ) ) {
+            $self->{'seen_helo_name'} = $helo_host;
 
             my $callbacks = $self->get_callbacks( 'helo' );
             foreach my $handler ( @$callbacks ) {
@@ -481,7 +494,7 @@ sub top_helo_callback {
             }
         }
         else {
-            $self->dbgout('Multiple HELO callbacks detected and ignored', $self->{'helo_name'} . ' / ' . $helo_host, LOG_DEBUG );
+            $self->dbgout('Multiple HELO callbacks detected and ignored', $self->{'seen_helo_name'} . ' / ' . $helo_host, LOG_DEBUG );
         }
 
         $self->set_alarm(0);
@@ -925,6 +938,7 @@ sub top_close_callback {
         $self->tempfail_on_error();
     }
     delete $self->{'helo_name'};
+    delete $self->{'seen_helo_name'};
     delete $self->{'raw_helo_name'};
     delete $self->{'c_auth_headers'};
     delete $self->{'auth_headers'};
@@ -2039,7 +2053,7 @@ Mail::Milter::Authentication::Handler
 
 =head1 VERSION
 
-version 2.20181130
+version 2.20181212
 
 =head1 DESCRIPTION
 
@@ -2328,10 +2342,18 @@ Manually check the current timeout, and throw if it has passed.
 
 Top level handler for handler setup.
 
+=head2 I<remap_connect_callback( $hostname, $ip )>
+
+Top level handler for the connect event for remapping only.
+
 =head2 I<top_connect_callback( $hostname, $ip )>
 
 Top level handler for the connect event.
 
+=head2 I<remap_helo_callback( $helo_host )>
+
+Top level handler for the HELO event for remapping only.
+
 =head2 I<top_helo_callback( $helo_host )>
 
 Top level handler for the HELO event.
